//****************************************************
//***** AITENDO FSTN SPI LCD controlling *************
//*************************** By Minarai Shokunin ****
//*************************** 2013/05/25          ****
//****************************************************

#include "p24FJ64GA002.h"
#include "oled.h"

extern unsigned char VirtualScreen[8][128];

// ************ prototypes ***************************
extern void OLED_init();
extern void WriteData(int,int);
extern void Copy2OLED(void);
extern void GR_FILL(unsigned char);
extern void GR_PSET(int,int,unsigned char);
extern void GR_LINE(int,int,int,int,unsigned char);
extern void GR_PUTCH(int,int,unsigned char, unsigned char);
extern void GR_PUTNUM(int,int,int, unsigned char);
extern void GR_PUTSTR(int,int,unsigned char *, unsigned char);
void game_screen_disp(void);
void game_round_init(void);
void wait_some(int);
void game_enemy_move(void);
void game_ship_move(void);
void game_missile_move(void);
void game_beam_move(void);
void game_add_beam(int);
void game_ufo_move(void);
void game_bgm(void);

// ************ CONFIGRATION BITS ********************
_CONFIG1(JTAGEN_OFF & GCP_OFF & GWRP_OFF & BKBUG_OFF & COE_OFF & ICS_PGx1 & FWDTEN_OFF )
_CONFIG2(IESO_OFF & FNOSC_FRCPLL & FCKSM_CSDCMD & OSCIOFNC_OFF & IOL1WAY_OFF & I2C1SEL_PRI & POSCMOD_NONE)


// ************ sound data ***************************
int sound_freq[] = {
	// C  , C# , D  , D# , E  , F  , F# , G  , G# , A  , A# , B 	
	  130 ,139 ,147 ,156 ,165 ,175 ,185 ,196 ,208 ,220 ,233 ,247,
	  261 ,277 ,294 ,311 ,330 ,350 ,370 ,392 ,415 ,440 ,466 ,494,
      523, 554 ,587 ,622 ,659 ,698 ,740 ,783 ,830 ,880 ,932 ,988,
      1046,1108,1175,1245,1319,1396,1480,1568,1661,1760,1865,1796,
      1,1,1,1,1,1,1,1,1,1,1,1
};
int sound_len[] = {
	2400,1200,600,300,150,		// 1, 1/2 , 1/4 , 1/8 , 1/16
    3600,1800,900,450,225,		// With Dot
	1600,800,400,200,100		// 3 renpu
};

const int sound_test[] = {
	24,1,26,1,28,1,29,1,31,1,33,1,35,1,
	0xffff,0xffff
};

const int sound_lost[] = {
	0,2,7,2,0,2,7,2,0,2,7,2,
	0xffff,0xffff
};

const int sound_shoot[] = {
	31+12,4,28+12,4,24+12,4,
	0xffff,0xffff
};

const int sound_hit[] = {
	24,4,28,4,31,4,
	0xffff,0xffff
};

const int sound_hitufo[] = {
	24,4,28,4,31,4,24,4,28,4,31,4,
	0xffff,0xffff
};

const int bgm[]= { 
	11+12,3,0xff,1,0xffff,0xffff,
    9+12,3,0xff,1,0xffff,0xffff,
	7+12,3,0xff,1,0xffff,0xffff,
	5+12,3,0xff,1,0xffff,0xffff
};

const int bgm2[]= { 
	24,4,28,4,24,4,28,4,0xffff,0xffff
};


int sound_enabled;			// Sound enable flag
int sound_left;				// left count for current tone
int sound_tone;				// current tone
int *sound_pointer;			// current tone/len data

int gl_round;				// round counter
int gl_ships_left;			// left ships counter
int gl_ship_x;				// ship left position
int gl_missile_f,gl_missile_x,gl_missile_y;				// missile info
int gl_enemy_x[24];										// enemy x,y
int gl_enemy_y[24];
int gl_enemy_f[24];
int gl_enemy_moving_dir;
int gl_enemy_moving_dir_pre;
int gl_enemy_moving_pnt;
int gl_enemy_moving_pnt_pre;
int gl_enemy_left;
int gl_beam_f[5],gl_beam_x[5],gl_beam_y[5];			// enemy beam info
int gl_shield_f[5],gl_shield_x[5],gl_shield_y[5];		// shield info
int gl_ufo_f,gl_ufo_x,gl_ufo_y;									// UFO info
int gl_score,gl_high;
int gl_ufo_count;
int gl_count;
int gl_lose;
int gl_cleared;
int gl_bgm;

// *********************************************
// ************ MAIN ***************************
// *********************************************

int main(void)
{
	int i;

	// ************ CPU Initialize **********

	_RCDIV = 0;							// PostScaler 32MHz
	TRISB = 0b0000011100000000;			// PORTB 15-11,7-0 = output  ; 10-9 = input
	CNPU1 = 0xffff;						// Set pull-ups
	CNPU2 = 0xffff;						// Set pull-ups

	// *********** Sound Itinialize ***************

	sound_enabled = NO;
	sound_left = 0;

	RPOR3bits.RP6R 		= 18;   // RP6 = OSC1
    PR2             	= 32000; 	// Timer2 period 
    T2CONbits.TCKPS 	= 0;    // Timer2 Pre scaler 1:1
    T2CONbits.TON   	= 0;    // Timer2 OFF
    OC1CONbits.OCTSEL 	= 0; 
    OC1CONbits.OCM    	= 5;      
    OC1R              	= PR2;   
    OC1RS             	= PR2/2; 

	// *********** Display Itinialize ***************

	OLED_init();
	GR_FILL(0);
	Copy2OLED();

	// *********** sound test ***************

 	PR1 = 4000; 
    T1CONbits.TCKPS 	= 2;    // Timer1 Pre scaler 1:64 = 16ms
    T1CONbits.TON   	= 1;    // Timer1 ON
 	IEC0bits.T1IE = 1;  // Timer1 interrupt enable
	sound_pointer = sound_test;
	//sound_enabled = YES;

	// *********** Game initial **************
	
	while(1) {

		gl_ships_left = 3;
		gl_high =  0;
		gl_score = 0;
		gl_round = 1;
		gl_bgm =0;

		for (i=0;i<3;i++) {
			GR_FILL(255);
			GR_PUTSTR(10,16,"SPACE INVADER",0);
			Copy2OLED();
			wait_some(1000);
			GR_FILL(0);
			GR_PUTSTR(10,16,"SPACE INVADER",1);
			Copy2OLED();
			wait_some(1000);
		}

		while (gl_ships_left >0) {	
			game_round_init();
			game_screen_disp();
			wait_some(5000);
			while ((gl_lose == NO)&&(gl_cleared == NO)) {
				game_enemy_move();
				game_ship_move();
				game_missile_move();
				game_beam_move();
				game_ufo_move();
				game_screen_disp();
				game_bgm();
				wait_some(150);
			}
			for (i=10;i<80;i++) {
				VirtualScreen[2][i]=0;
				VirtualScreen[3][i]=0;
				VirtualScreen[4][i]=0;
			}
			if (gl_cleared == YES) {
				GR_PUTSTR(10,16,"YOU WON!",1);
				Copy2OLED();
				wait_some(5000);
			}
			if ((gl_lose ==YES)&&(gl_ships_left >0)) {
				GR_PUTSTR(10,16,"YOU LOST.",1);
				GR_PUTSTR(10,24,"PUSH L",1);
				Copy2OLED();
				while (SW_L ==1) {};
				while (SW_L ==0) {};
				while (SW_L ==1) {};
				gl_ships_left -= 1;
			}
		}
		for (i=10;i<80;i++) {
			VirtualScreen[2][i]=0;
			VirtualScreen[3][i]=0;
			VirtualScreen[4][i]=0;
		}
		GR_PUTSTR(10,16,"GAME OVER",1);
		GR_PUTSTR(10,24,"PUSH L",1);
		if (gl_score > gl_high) {
			gl_high = gl_score;
		}
		Copy2OLED();
		while (SW_L ==1) {};
		while (SW_L ==0) {};
		while (SW_L ==1) {};

	}
	
}


// *********************************************
// ************ Sub routines    ****************
// *********************************************
void game_bgm(void) {
	if (sound_enabled == NO) {
		if (gl_ufo_f ==YES) {
			sound_pointer = bgm2;
			sound_enabled = YES;
		}else{
			sound_pointer = bgm + (gl_bgm *6);
			sound_enabled = YES;
			gl_bgm = (gl_bgm+1)%4;
		}
	}
}


void game_ufo_move(void) {
	if (gl_ufo_f ==0) {
		gl_ufo_count++;
		if (gl_ufo_count > 100) {
			gl_ufo_count =0;
			gl_ufo_f =1 ;
			gl_ufo_y = 0;
			gl_ufo_x = 70;
		}
	}else{
	   gl_ufo_x -=2;
	   if (gl_ufo_x <4) {
		 gl_ufo_f =0;
	   }else{
		 GR_PUTCH(gl_ufo_x,gl_ufo_y,0x68,1);
	   }
	}
}
void game_ship_move(void) {
	if (SW_L ==0) {
		if (gl_ship_x >=6) {
			gl_ship_x -=4;
		}
	}
	if (SW_R ==0) {
		if (gl_ship_x <=80) {
			gl_ship_x +=4;
		}
	}
	if (SW_S ==0) {
		if (gl_missile_f ==0) {
			gl_missile_f =1;
			gl_missile_x = gl_ship_x;
			gl_missile_y = 54;
			sound_pointer = sound_shoot;
			sound_enabled = YES;
		}
	}
}

void game_beam_move(void) {

	int i,j,dx,dy;

	for (i=0;i<3;i++) {
		if (gl_beam_f[i] >0) {
			gl_beam_y[i] += 4;
			if (gl_beam_y[i] > 54) {
				gl_beam_f[i] =0;
			}
		}
		
		// ******* Shild hit check ******
		for (j=0;j<4;j++) {
			if (gl_shield_y[j] > gl_beam_y[i]) {
				dy = gl_shield_y[j] - gl_beam_y[i];
			}else{
				dy =  gl_beam_y[i] - gl_shield_y[j];
			}
			if (gl_shield_x[j] > gl_beam_x[i]) {
				dx = gl_shield_x[j] - gl_beam_x[i];
			}else{
				dx =  gl_beam_x[i] - gl_shield_x[j];
			}
			// ******* HIT Shield *******
			if ((gl_beam_f[i]>0)&&(gl_shield_f[j]>0)&&(dx < 4)&&(dy<4)) {
				gl_shield_f[j] += 1;
				gl_beam_f[i] =0;
			}
		}
		// ******* Ship hit Check **********

		if (56 > gl_beam_y[i]) {
			dy = 56 - gl_beam_y[i];
		}else{
			dy =  gl_beam_y[i] - 56;
		}
		if (gl_ship_x > gl_beam_x[i]) {
			dx = gl_ship_x - gl_beam_x[i];
		}else{
			dx =  gl_beam_x[i] - gl_ship_x;
		}
		// ******* HIT Ship *******
		if ((gl_beam_f[i]>0)&&(dx < 4)&&(dy<4)) {
			gl_lose = YES;
			gl_beam_f[i] =0;
			sound_pointer = sound_lost;
			sound_enabled = YES;
		}
	}
}
void game_missile_move(void) {
	int i,dx,dy;

	if (gl_missile_f >0) {
		gl_missile_y -=6;
		if (gl_missile_y < 4) {
			gl_missile_f =0;
		}

		// ******* UFO hit check ******
		dy =  gl_missile_y - gl_ufo_y;
		if (gl_ufo_x > gl_missile_x) {
			dx = gl_ufo_x - gl_missile_x;
		}else{
			dx =  gl_missile_x - gl_ufo_x;
		}
		// ******* HIT UFO*******
		if ((gl_ufo_f == 1)&&(dx < 8)&&(dy<8)) {
				gl_missile_f =0;
				gl_score += 50;
				gl_ufo_f =2;
				sound_pointer = sound_hitufo;
				sound_enabled = YES;
		}
		// ******* Shild hit check ******
		for (i=0;i<4;i++) {
			if (gl_shield_y[i] > gl_missile_y) {
				dy = gl_shield_y[i] - gl_missile_y;
			}else{
				dy =  gl_missile_y - gl_shield_y[i];
			}
			if (gl_shield_x[i] > gl_missile_x) {
				dx = gl_shield_x[i] - gl_missile_x;
			}else{
				dx =  gl_missile_x - gl_shield_x[i];
			}
			// ******* HIT Shield *******
			if ((gl_shield_f[i]>0)&&(dx < 4)&&(dy<4)) {
				gl_shield_f[i] += 1;
				gl_missile_f =0;
			}
		}

		// ******* Enemy hit check ******
		for (i=0;i<18;i++) {
			if (gl_enemy_y[i] > gl_missile_y) {
				dy = gl_enemy_y[i] - gl_missile_y;
			}else{
				dy =  gl_missile_y - gl_enemy_y[i];
			}
			if (gl_enemy_x[i] > gl_missile_x) {
				dx = gl_enemy_x[i] - gl_missile_x;
			}else{
				dx =  gl_missile_x - gl_enemy_x[i];
			}
			// ******* HIT ENEMY *******
			if ((gl_enemy_f[i]>0)&&(dx < 4)&&(dy<4)) {
				gl_enemy_f[i]=255;
				gl_score +=10;
				gl_missile_f =0;
				sound_pointer = sound_hit;
				sound_enabled = YES;
				gl_enemy_left --;
				if (gl_enemy_left == 0) {
					gl_cleared = YES;
				}
			}
		}
	}
}


void game_enemy_move(void) {
	int done,i,j,dx,dy;

	done = NO;	

	for (i=0;i<18;i++) {
		if (gl_enemy_f[i] > 0) {
			game_add_beam(i);
		}	
		// ******* Shild hit check ******
		for (j=0;j<4;j++) {
			if (gl_shield_y[j] > gl_enemy_y[i]) {
				dy = gl_shield_y[j] - gl_enemy_y[i];
			}else{
				dy =  gl_enemy_y[i] - gl_shield_y[j];
			}
			if (gl_shield_x[j] > gl_enemy_x[i]) {
				dx = gl_shield_x[j] - gl_enemy_x[i];
			}else{
				dx =  gl_enemy_x[i] - gl_shield_x[j];
			}
			// ******* HIT Shield *******
			if ((gl_enemy_f[j]>0)&&(dx < 4)&&(dy<8)) {
				gl_shield_f[j] = 0;
			}
		}
	}

	while (done == NO) {
		if (gl_enemy_moving_dir == 1) {
			if (gl_enemy_moving_pnt == -1) {
				if (gl_enemy_moving_dir != gl_enemy_moving_dir_pre) {
					for (i=0;i<18;i++) {
						gl_enemy_y[i] +=4;
						if ((gl_enemy_f[i]>0)&&(gl_enemy_y[i]>46)) {
							gl_lose = YES;
							sound_pointer = sound_lost;
							sound_enabled = YES;
						}
					}
				}
				gl_enemy_moving_dir = gl_enemy_moving_dir_pre;
				gl_enemy_moving_pnt = gl_enemy_moving_pnt_pre;

			}else{
				if (gl_enemy_f[gl_enemy_moving_pnt] >0) {
					gl_enemy_x[gl_enemy_moving_pnt] += 4;
					done = YES;
					if (gl_enemy_x[gl_enemy_moving_pnt]>=80) {
						gl_enemy_moving_dir_pre = -1;
						gl_enemy_moving_pnt_pre = 0;
					}
				}
				gl_enemy_moving_pnt -= 1;
			}
		}else{
			if (gl_enemy_moving_pnt == 18) {
				if (gl_enemy_moving_dir != gl_enemy_moving_dir_pre) {
					for (i=0;i<18;i++) {
						gl_enemy_y[i] +=4;
						if ((gl_enemy_f[i]>0)&&(gl_enemy_y[i]>46)) {
							gl_lose = YES;
							sound_pointer = sound_lost;
							sound_enabled = YES;
						}
					}
				}
				gl_enemy_moving_dir = gl_enemy_moving_dir_pre;
				gl_enemy_moving_pnt = gl_enemy_moving_pnt_pre;
			}else{
				if (gl_enemy_f[gl_enemy_moving_pnt] >0) {
					gl_enemy_x[gl_enemy_moving_pnt] -= 4;
					done = YES;
					if (gl_enemy_x[gl_enemy_moving_pnt]<=6) {
						gl_enemy_moving_dir_pre = 1;
						gl_enemy_moving_pnt_pre = 17;
					}
				}
				gl_enemy_moving_pnt += 1;
			}
		}
	}
}

void game_add_beam(int enemy) {

	int j,v;	

	// ********** Beam adding ******
	v=0;

	if ((rand(0)%100)==1) {	
		for (j=0;j<3;j++) {
			if ((gl_beam_f[j] == 0)&&(v==0)) {
				gl_beam_f[j]=1;
				gl_beam_x[j]=gl_enemy_x[enemy];
				gl_beam_y[j]=gl_enemy_y[enemy]+4;
				v=1;
			}
		}
	}	
}

void game_screen_disp(void){
	int i,j;

	GR_FILL(0);
	GR_PUTSTR(96,0,"HI",1);
	GR_PUTNUM(96,8,gl_high,1);
	GR_PUTSTR(96,16,"SC",1);
	GR_PUTNUM(96,24,gl_score,1);
	for (i=0;i<gl_ships_left;i++) {
		GR_PUTCH(96+i*10,40,0x60,1);
	}
	//*********** Invaders **********
	for (i=0;i<18;i++) {
		if (gl_enemy_f[i]>0) {
			if (gl_enemy_f[i] == 255) {
				GR_PUTCH(gl_enemy_x[i],gl_enemy_y[i],0x67,1);
				gl_enemy_f[i] = 0;
			}else{
				GR_PUTCH(gl_enemy_x[i],gl_enemy_y[i],0x67+gl_enemy_f[i]*2+gl_count,1);
			}
		}
	}
	//********** Beams **************
	for (i=0;i<3;i++) {
		if (gl_beam_f[i]>0) {
			GR_PUTCH(gl_beam_x[i],gl_beam_y[i],0x62,1);			
		}
	}
	gl_count = 1 - gl_count;

	//********** Shields **************
	for (i=0;i<4;i++) {
		if (gl_shield_f[i]>0) {
			GR_PUTCH(gl_shield_x[i],gl_shield_y[i],gl_shield_f[i]/2+0x63,1);
			if 	(gl_shield_f[i]>5) {
				gl_shield_f[i]=0;
			}		
		}
	}

	//********** Ship ***************
	GR_PUTCH(gl_ship_x,56,0x60,1);	

	//********** Missile ************
	if (gl_missile_f > 0) {
		GR_PUTCH(gl_missile_x,gl_missile_y,0x61,1);	
	}
	//********** UFO ************
	if (gl_ufo_f ==1 ) {
		GR_PUTCH(gl_ufo_x,0,0x68,1);	
	}
	if (gl_ufo_f ==2 ) {
		GR_PUTSTR(gl_ufo_x,0,"50",1);
		gl_ufo_f =0;	
	}

	Copy2OLED();
}


void game_round_init(void) {
	
	int i,j;

	gl_ship_x		= 4;
	gl_missile_f	= 0;
	gl_lose 		= NO;

	for (i=0;i<3;i++) {
		for (j=0;j<6;j++) {
			gl_enemy_f[i*6+j] = i + 1;
			gl_enemy_x[i*6+j] = 10+j*10;
			gl_enemy_y[i*6+j] = 10+i*10;
		}
	}

	for (i=0;i<4;i++) {
		gl_shield_f[i] = 1;
		gl_shield_x[i] = 14+20*i;
		gl_shield_y[i] = 46;
	}

	for (i=0;i<3;i++) {
		gl_beam_f[i] = 0;
	}
	gl_ufo_f = 0;
	gl_ufo_count = 0;
	gl_count = 0;
	gl_enemy_moving_pnt = 17;
	gl_enemy_moving_dir = 1;
	gl_enemy_moving_pnt_pre = 17;
	gl_enemy_moving_dir_pre = 1;
	gl_enemy_left = 18;
	gl_cleared = NO;
}


//********** Simple wait loop *************
void wait_some(int t)
{
	int i,j;
	for (i=0;i<t;i++) {
		for (j=0;j<1000;j++) {
		}
	}
}

// *********************************************
// ************ Timer Interrupt ****************
// *********************************************

void __attribute__((interrupt, auto_psv)) 
_T1Interrupt(void)
{
	unsigned long wk;

    IFS0bits.T1IF = 0; 		// Timer 1 Interrupt flag clear

	if (sound_enabled == NO) {
		return;
	}

	if (sound_left == 0) {

		if (*sound_pointer == 0xffff) {
			sound_enabled = 0;
    		T2CONbits.TON   	= 0;    // Timer2 OFF
		}else{
			if (*sound_pointer == 0xff) {
				sound_tone = sound_freq[*sound_pointer];
				sound_pointer++;
				sound_left = sound_len[*sound_pointer]/50;
				sound_pointer++;
    			T2CONbits.TON   	= 0;    // Timer2 OFF
			}else{
				sound_tone = sound_freq[*sound_pointer];
				sound_pointer++;
				sound_left = sound_len[*sound_pointer]/50;
				sound_pointer++;
				wk = ((unsigned long)(16000000))/((unsigned long)sound_tone);
				PR2 = (int)wk; 
    			OC1R              	= PR2;   
    			OC1RS             	= PR2/2;
    			T2CONbits.TON   	= 1;    // Timer2 ON
			}
		}

	}else{
		sound_left--;
	}

}

