//****************************************************
//***** AITENDO 128x64 OLED Graphic routines  ********
//*************************** By Minarai Shokunin ****
//*************************** 2013/09/21          ****
//****************************************************

#include "p24FJ64GA002.h"
#include "oled.h"
#include "font8x8.h"

extern const unsigned char font8x8[];
extern unsigned char VirtualScreen[8][128];

// ************ prototypes ***************************
void GR_FILL(unsigned char);
void GR_PSET(int,int,unsigned char);
void GR_PUTCH(int,int,unsigned char,unsigned char);
void GR_LINE(int,int,int,int,unsigned char);
void GR_PUTCH(int,int,unsigned char, unsigned char);
void GR_PUTSTR(int,int,unsigned char *, unsigned char);

// ************ Values ***************************
const unsigned char bits[] = {1,2,4,8,16,32,64,128};


// ************ Initialization ***************************


//********** Fill data *************
void GR_FILL(unsigned char dt) {
	int i,j;

	for (i=0;i<8;i++) {
		for (j=0;j<128;j++) {
			VirtualScreen[i][j] = dt;
		}
	}
}

//********** Put string *************
void GR_PUTSTR(int x,int y,unsigned char *dt, unsigned char col) {
	unsigned char *p0;
	int x0;	
	
	p0 = dt;
	x0 = x;

	while (*p0 != 0x00) {
		GR_PUTCH(x0,y,*p0++,col);
		x0 += 8;
	}	
}

//********** Put Number *************
void GR_PUTNUM(int x,int y,int v, unsigned char col) {
	unsigned char v0;
	int i,x0,vv;
	
	x0 = x+3*8;
	vv=v;

	for (i=0;i<4;i++) {
		GR_PUTCH(x0,y,0x30+(vv%10),col);
		x0 -= 8;
		vv /=10;
	}
}



//********** Put character *************

void GR_PUTCH(int x,int y,unsigned char dt, unsigned char col) {

	int y0,y1,i,p0;
	unsigned char c0,c1;
		
	y0=y/8; y1=y & 0x0007;
	p0 = (dt - 0x20) * 8;

	if (y1 == 0) {
		for (i=0;i<8;i++) {
			if (col == 0) {
				VirtualScreen[y0][x+i] &= 255-font8x8[p0+i];
			}else{
				VirtualScreen[y0][x+i] |= font8x8[p0+i];
			}
		}
	}else{
		for (i=0;i<8;i++) {
			c0 = font8x8[p0+i]<<y1;
			c1 = font8x8[p0+i]>>(8-y1);
			if (col == 0) {
				VirtualScreen[y0][x+i] &= 255 - c0;
				VirtualScreen[y0+1][x+i] &= 255 - c1;
			}else{
				VirtualScreen[y0][x+i] |= c0;
				VirtualScreen[y0+1][x+i] |= c1;
			}
		}
	}
}


//********** Pset *************
void GR_PSET(int x,int y,unsigned char dt) {
	int y0,y1;
	unsigned char c0,c1;

	y0=y/8; y1=y & 0x0007;
	c0 = VirtualScreen[y0][x];
	c1 = bits[y1];

	if (dt == 0) {
		 VirtualScreen[y0][x] = c0 & (255-c1);	
	}else{
		 VirtualScreen[y0][x] = c0 | c1;	
	}
}

//********** Line *************

void GR_LINE(int x0,int y0,int x1,int y1,unsigned char col)
{
  int i,dx,dy,e,sx,sy,xx0,yy0;

  xx0=x0;
  yy0=y0;

  if (x1 > x0) {
		dx = x1-x0;
		sx = 1;
  }else{
		dx = x0-x1;
		sx = -1;
  }

  if (y1 > y0) {
		dy = y1-y0;
		sy = 1;
  }else{
		dy = y0-y1;
		sy = -1;
  }

  if ( dx > dy ) {
    e = -dx;
    for ( i = 0 ; i <= dx ; i++ ) {
      GR_PSET(xx0,yy0, col );
      xx0 += sx;
      e += 2 * dy;
      if ( e >= 0 ) {
        yy0 += sy;
        e -= 2 * dx;
      }
    }
  } else {
    e = -dy;
    for ( i = 0 ; i <= dy ; i++ ) {
      GR_PSET(xx0,yy0, col );
      yy0 += sy;
      e += 2 * dx;
      if ( e >= 0 ) {
        xx0 += sx;
        e -= 2 * dy;
      }
    }
  }
}
